# encoding: utf-8
"""Write Django model properties that can be used in database queries."""

from __future__ import unicode_literals

__version__ = '1.4.0'
__author__ = 'Marcus Klöpfel'
__copyright__ = 'Copyright 2020, Marcus Klöpfel'
__license__ = 'BSD'
__maintainer__ = 'Marcus Klöpfel'
__email__ = 'marcus.kloepfel@gmail.com'
__status__ = 'Production/Stable'
